# Flask-SQLAlchemy-HAL Support

The goal of this project is to trivialize the creation of a HAL/JSON REST
api from a SQLAlchemy declarative model using the Flask framework.  The resulting
API is secured using JWT.  Docstrings are automatically exported as HAL
documentation.


#### QuickStart

```
from hyperdns.flask import web

@web.api({
	'root':(name,path)
})
class MyModelClass(app.db.Model):
	app.db.Column
	app.db.Map
	app.db.Scalar
	
	
web.init_app(app)

```

## Data Types

The module provides three vehicles for exposing data via HAL - properties
are used to manage primitive elements, scalars are non-collection HAL entities,
and maps are HAL collections.  These are extensions of corresponding SQLAlchemy
types, and almost all of the functionality is passed on to the corresponding
SQLAlchemy or Flask-SQLAlchemy element after stripping out the specific keyword
arguments that control HAL exposure.

#### Properties - app.db.Column

Properties are automatically discovered and exposed.  Visibility is controlled
by prefixing the property name with an '_' to hide the property from the automatic
exposure.  Properties must be primitive types - meaning that they must be JSON
serializable but not HAL objects.  Simple JSON structures, strings, numbers, booleans
and that sort of thing are the spice that properties are made of.

#### Scalars - app.db.Scalar

Scalars control the embedding of singular objects that are slightly more complicated
than properties but which do not form a collection.

#### Maps - app.db.Map

Maps are the single type of collection and a map declaration must identify an
embedded type as well as an index property.  Maps are not collections of
primitives, but collections of HAL rendered objects which may be embedded.


	class myClass(app.db.Model):
	
		app.db.Map('AnotherClass','akey',....)

This only captures the first two arguments, the implementing class of the collection
elements and the key used to index them - the rest of the arguments are passed up to
Flask-SQLAlchemy's app.db.relationship()

# Registering classes and Identifying Roots

HAL apis require the publication of a single endpoint, the API Root.  Once
a potential consumer can access the API root, the entire API can be navigated,
explored, and understood.  When defining your model, use the @web.api() method

```
@web.api()
class myClass(app.db.Model):
	...
```


## How to define a root

The @web.api method takes a map as an argument.  This allows you to configure the
way in which the class is exposed and whether or not it is a root object.  When declaring
a root object you need to identify the name of the root to use, and the path of the
root relative to the entire applications path.

For example, if you had
```
@web.api({
	'root':('account','/my')
})
class myClass(app.db.Model):
	...
```

and you later called @web.instrument('http://localhost:5000/somepath') then your HAL
api would be available at 'http://localhost:5000/somepath/my' and all objects (and thus)
uris would be relative to the 'account' root object.

## Where do the roots come from



# Identity and Security (auth.py)

The @web.auth

## Configuring Security

```
@web.auth.config({
    DISABLE_AUTH=False
    CLIENT_BASE="http://localhost:5000"
    CLIENT_ID = "EIOlHBKWUyavVotmpkb6r3MT1sHCACrW"
    CLIENT_SECRET = "Kt00Up7IAdbQZ44Ops_WMhm9tVj_tIXui0Me7nIU5VdEEO2x7MA5dVGx5S1yN2CO"
    INTERNAL_SECRET = "Kt00Up7IAdbQZ44Ops_WMhm9tVj_tIXui0Me7nIU5VdEEO2x7MA5dVGx5S1yN2CO"
})
```

## Loading Users

The application is responsible for identifying a user-loader which takes the JWT payload
and returns the appropriate user object that will be returned by web.current_user.

@web.auth.user_loader

## Internal Keys

internal keys are used to support agents operating on behalf of users or on behalf of the
system.

## SSL

We require access via SSL - no HTTP access is allowed.  All http access is
redirected to https.  This is implemented, currently, in __init__.py, and it
simply issues a 302 redirect if the scheme of the inbound URL is http.  The
request is redirected to the same URI via https


## CORS

auth.py provides an annotation "crossdomain" which supports CORS

## Identity

### User

web.current_user
web.current_account

### Account

## security
Security is managed by JWT or JavaScript Web Tokens.

### JWT claims

List the claims....


# API details

## JWTError

issued when?


## Returning Results

### Compressing the results

### returning HTTP non-200 responses

The web.result module is 

### Returning HAL

### Returning raw JSON

### returning plaintext


## How the instrumentation works

1. the Flask object is created elsewhere and web.init_app(app) is called
	- the app.db.Map and app.db.Scalar objects are created
	- the definition.registry and api_root.registry maps are initialized, these
	  contain the classes that need to be exposed.
2. the model is built using @web.api annotations and Map and Scalar objects
	- this will register all of the exposable classes in definition.registry
	  and if @web.api(root='/path') is provided, it will also add the object
	  to api_root.registry()
3. you call web.instrument(baseUri) prior to starting your app

instrument 