"""Expose Flask-SQLAlchemy models as HAL objects.

Basic Usage:
    1. import hyperdns.flask as web
    2. web.current_user is the current user

"""
from werkzeug.local import LocalProxy,LocalStack
from urllib.parse import urlparse,urlunparse
from flask import _request_ctx_stack,request,Response,redirect
from sqlalchemy.orm.collections import attribute_mapped_collection

# make the user and account available - these come from auth.py ultimately
# but we define them here - @todo - is this the right place?
current_user = LocalProxy(lambda: _request_ctx_stack.top.current_user)
hyperdns_context = LocalProxy(lambda: _request_ctx_stack.top.hyperdns)

# local reference to 'the app object' - there has *got* to be a better way to do this
_app=None
    
def init_app(app):
    """Initializes components of the flask module that require access to
    the actual app.
    
     - create /angular/directives.js to expose the vendor directives
       NOTE: I think this belongs in the vendor modules
       
     - redirect all http:// to https://
     
     - initialize the app.db.Map and app.db.Scalar elements.
    """
    global _app
    _app=app

    # @todo - does this belong here?
    # exposes
    import hyperdns.vendor
    @app.route('/angular/directives.js',methods=['GET'])
    def return_angular_directives():
        """Returns the directives for the set of available vendors
        """
        raw=hyperdns.vendor.DriverManager.directives
        amd="define('vendors',['angular'],function() {"+raw+"})"
        return Response(amd,mimetype='application/javascript')

    # require SSL
    @staticmethod
    @app.before_request
    def redirect_to_ssl():
        """Redirect all requests to SSL"""
        return
        urlparts = urlparse(request.url)
        urlparts_list = list(urlparts)
        urlparts_list[0] = 'https'
        return redirect(urlunparse(urlparts_list), code=302)

    # record the implementation type and index key for child maps
    def __map(class_name,index_key,**kwargs):
        """Capture the first two args, otherwise the same as app.db.relationship() """

        if 'collection_class' in kwargs:
            result=app.db.relationship(class_name,**kwargs)
        else:
            result=app.db.relationship(class_name,
                collection_class=attribute_mapped_collection(index_key),**kwargs)

        if 'backref' not in kwargs:
            raise Exception('HAL collections need backref to determine parent for link generation - failure building map of %s' % class_name)
        result.parent_field=kwargs['backref']
        
        result.impl_type=class_name
        result.impl_index=index_key

        return result
    app.db.Map=__map
        
    def __scalar(class_name,**kwargs):
        result=app.db.relationship(class_name,uselist=False,**kwargs)
        if 'backref' not in kwargs:
            raise Exception('HAL scalars need backref to determine parent for link generation - failure building map of %s' % class_name)
        result.parent_field=kwargs['backref']

        result.impl_type=class_name
        result.impl_index=None
        return result
    app.db.Scalar=__scalar


# huh......      
class db_commit(object):
    """
    for some reason we're getting automatic ROLLBACK - on top of that, pycharm
    is not allowing me to set breakpoints
    """
    def __init__(self):
        pass
    
    def __call__(self,f):
        def _db_commit(*args, **kwargs):
            try:
                result=f(*args, **kwargs)
                _app.db.session.commit()
            except:
                _app.db.session.rollback()
                raise
            return result
        d=_db_commit
        d.__name__=f.__name__
        return d
        
# these 
class JWTError(Exception):
    def __init__(self, error, description, status_code=400, headers=None):
        self.error = error
        self.description = description
        self.status_code = status_code
        self.headers = headers
# currently unused.... was once, retaining as a placeholder - may implement
# via redis and/or here, and if so will move to rate.py or something like that.
class rate_limit(object):
    """Rate limiting
    
    To prevent abuse, it is standard practice to add some sort of rate limiting to an
    API. RFC 6585 introduced a HTTP status code 429 Too Many Requests to accommodate
    this.
    
    X-Rate-Limit-Limit - The number of allowed requests in the current period
    X-Rate-Limit-Remaining - The number of remaining requests in the current period
    X-Rate-Limit-Reset - The number of seconds left in the current period
    """
    def __init__(self,contentType='application/json'):
        pass
    
    def __call__(self,f):
        @wraps(f)
        def decorated(*args,**kwargs):
            return f(*args,**kwargs)
        return decorated

# now, pull in the sub-modules
from .api import api
from .auth import auth,crossdomain
from .result import result

