"""HAL API annotations for Flask-SQLAlchemy (python3 only).


Basic Usage
    @web.api()
    class myClass(app.db.Model):
    
        an_index=app.db.Column(Integer,primary_key)
        sub_object=app.db.Scalar('myClass')
        a_collection=app.db.Map('myClass','an_index')

    this will generate a 
"""
import inspect
import traceback
import hyperdns.hal.render as HAL
import hyperdns.flask
from hyperdns.flask import current_user,hyperdns_context
from .result import result
from functools import wraps
from flask import request,Response
import sys
import inspect
from  sqlalchemy.orm.attributes import (
    InstrumentedAttribute,
    CollectionAttributeImpl,
    ScalarAttributeImpl
    )
from sqlalchemy.orm.relationships import RelationshipProperty
from urllib.parse import urlparse


#
class definition(object):
    """the @web.api annotations capture the classes - each class has an
    api definition, each one, as it is encountered
    
    Expected Usage:
        definition.registry()
        
    """
    registry={}
    
    def __init__(self,spec,class_):
        self.spec=spec
        self.class_=class_
        self.name=class_.__name__
        assert self.registry.get(self.name)==None
        self.registry[self.name]=self
        self.parents={}

    def __repr__(self):
        return "Defn:%s" % (self.name)
            
#
class api_root(definition):
    """A definition with the 'root' qualifier
    """
    roots={}
    
    def __init__(self,spec,class_):
        super(api_root,self).__init__(spec,class_)
        rootspec=self.spec.get('root')
        assert rootspec!=None
        (self.root_name,self.root_path)=rootspec
        self.roots[self.name]=self
 
    def __repr__(self):
        return "Root:%s" % (self.name)

#
class path_processor(object):
    """path processors are responsible for establishing the object context
    due to the flask path.
    """
    def __init__(self,uri,path,depth=0):
        """record this path relative to a root
        """
        self.root_name=path[:1][0]
        self.uri=uri
        self.depth=depth
        self.object_path=path[1:]
    
    def __repr__(self):
        newpath=self.object_path[:]
        newpath.insert(0,self.root_name)
        return "PP(%d,%s,%s)" % (self.depth,newpath,self.uri)
        
    def for_action(self,arg):
        newpath=self.object_path[:]
        newpath.insert(0,self.root_name)
        uri='%s%s' % (self.uri,arg)
        return path_processor(uri,newpath,self.depth)
        
    def for_child(self,arg):
        newpath=self.object_path[:]
        newpath.insert(0,self.root_name)
        newpath.append(arg)
        uri='%s/%s' % (self.uri,arg)
        return path_processor(uri,newpath,self.depth+1)

    def for_collection(self,c_name):
        newpath=self.object_path[:]
        newpath.insert(0,self.root_name)
        newpath.append(c_name)
        uri='%s/%s/<string:arg_%d>' % (self.uri,c_name,self.depth+1)
        return path_processor(uri,newpath,self.depth+1)
            
    def setup_context(self,kwargs,request):
        """ set request.hyperdns['context']=web path

        example Path
            ['account','zones','resources']
        is equivalent to
            request.account.zones.get('my.zone.').resources.get('name1')
            
        request.hyperdns['context'] will be the path
        request.hyperdns['parent_context'] will be the last resolved object
            prior to this one, so if the context object is None, then
            you can get to the parent_context and use that to switch.
        """
        path_arguments=kwargs
        root_obj=hyperdns_context.get('available_roots',{}).get(self.root_name)
        cursor=root_obj
        parent_obj=None
        collection=None
        path=[]
        key=None
        scalar_name=None
        collection_name=None
        
        #print("Looking for",self.object_path)
        for path_elt in self.object_path:
            path.append(cursor)
            collection=getattr(cursor,path_elt,None)

            keyname='arg_%s' % (len(path))
            if keyname in path_arguments:
                key=path_arguments[keyname]
                parent_obj=cursor
                #print("deref",key,"in",collection)
                cursor=collection.get(key)
                collection_name=path_elt
            else:
                parent_obj=cursor
                cursor=collection
                collection=None
                scalar_name=path_elt
                    
        hyperdns_context['context']=cursor
        hyperdns_context['last_key']=key
        hyperdns_context['collection']=collection
        hyperdns_context['collection_name']=collection_name
        hyperdns_context['scalar_name']=scalar_name
        hyperdns_context['parent_obj']=parent_obj
        hyperdns_context['obj_path']=path
        hyperdns_context['root_obj']=root_obj
        hyperdns_context['root_scope']=self.root_name
        class context(dict):
            def __init__(self,kwargs):
                self.update(hyperdns_context)
                self.user=hyperdns_context['user']
                self.input=request.json
                self.object=cursor
                self.obj_path=path
                self.args=kwargs
                self.method=request.method
        #print("KEY",hyperdns_context['last_key'])
        return context(kwargs)
        
#
class api(object):
    """Indicate that a class is to be exposed, optionally determine
       if it has a root.
    
    Example:
        @web.api()
    """
    #             
    def __init__(self,spec={}):
        self.spec=spec
    
    def __call__(self,cls):
        """This populates the registry with unresolved routes
        """        
        if self.spec.get('root')!=None:
            d=api_root(self.spec,cls)
        else:
            d=definition(self.spec,cls)
        return cls

    class action(object):
        """This exposes a POST action on an object.  The object will be located
        and the method called with the JSON data from the request passed as the
        second argument.
    
        Example:
            @action('/settings/ignore')
            def a_method(self,input)
    
        """
        def __init__(self,uri,method='POST',hal=False):
            self.uri=uri
            self.method=method
            self.hal=hal
    
        def __call__(self,f):
            f.action={
                'func':f,
                'doc':f.__doc__,
                'uri':self.uri,
                'returns-hal':self.hal,
                'method':self.method
            }
            return f


    @classmethod
    def _analyze_model(api):
        """Look through all the classes that have been registered and, for each
            class, determine the properties, scalars, and maps.  Add those to the
            definition associated with that class.  This method does not look
            at the connection between classes, just the construction of each
            class.
        """
        for defn_key in sorted(definition.registry.keys()):
            defn=definition.registry[defn_key]
            spec=defn.spec
            cls=defn.class_
            boring = dir(type('dummy', (object,), {}))
            attrs=[item
                    for item in inspect.getmembers(cls)
                    if item[0] not in boring
                    and not item[0].startswith('_')
                    and isinstance(item[1],InstrumentedAttribute) ]
            defn.properties=[item
                    for item in attrs
                    if isinstance(item[1].impl,ScalarAttributeImpl)
                    and not isinstance(item[1].property,RelationshipProperty)]
            _scalars=[(item[0],
                    definition.registry.get(
                        getattr(item[1].property,'impl_type',None)))
                    for item in attrs
                    if isinstance(item[1].impl,ScalarAttributeImpl)
                    and isinstance(item[1].property,RelationshipProperty)]

            defn.scalars=[(s[0],s[1]) for s in _scalars if s[1] is not None]
            defn.collections=[(item[1].property.key,
                    definition.registry.get(
                        getattr(item[1].property,'impl_type',None)),
                    getattr(item[1].property,'impl_index',None))
                    for item in attrs
                    if isinstance(item[1].impl,CollectionAttributeImpl)]
            
            defn.actions=[item.action
                    for n,item in inspect.getmembers(cls,inspect.isfunction)
                    if hasattr(item,'action') ]
            
            defn.create=None
            if hasattr(cls,'create'):
                defn.create=getattr(cls,'create')
        
            defn.delete=None
            if hasattr(cls,'delete'):
                defn.delete=getattr(cls,'delete')

            defn.update=None
            if hasattr(cls,'update'):
                defn.update=getattr(cls,'update')
        
            
    @classmethod
    def _map_graph(api,defn,pp):
        """travel the graph from this element down until we either stop or locate
        a cycle.  each visit occurs within the scope of a root element, forming
        multiple trees as we flatten the graph out.
        """
        api._instrument_from_defn_with_pp(defn,pp)
                
        for (c_name,c_defn,c_index) in defn.collections:
            api._instrument_collection(defn,pp.for_child(c_name),c_name,c_index)            
            api._map_graph(c_defn,pp.for_collection(c_name))

        for (s_name,s_defn) in defn.scalars:
            api._map_graph(s_defn,pp.for_child(s_name))
            
    @classmethod
    def instrument(api,baseUri=None):
        """Determine what needs to be exposed and then expose it.
        """
        api._analyze_model()

        # now, scan the graph from each root, doing a depth first traversal and
        # setting up an empty array of all the different base uris for each root
        for root in api_root.roots.values():
            uri="%s%s" % (baseUri,root.root_path)
            api._map_graph(root,path_processor(uri,[root.root_name]))
                
        # add HAL support
        for defn in definition.registry.values():
            HAL.attach_HAL_support(HAL.hal_spec(
                defn.class_,
                scalars=defn.scalars,
                collections=defn.collections,
                properties=defn.properties
            ))
                
        return

            
    @classmethod
    def _instrument_collection(api,defn,pp,c_name,c_index):
        def read_collection(context):
            obj=hyperdns_context.get('context')
            if obj==None:
                return result.not_found()
            return HAL.hal_collection(
                data=obj.values(),parent=obj,name=c_name,keyfield=c_index
                )
        api._attach_method(defn,pp,'GET',read_collection)

    @classmethod
    def _instrument_action(api,defn,pp,action):
        def execute_action(context):
            method=action['func']
            obj=hyperdns_context.get('context')
            if obj==None:
                return result.not_found()
            return method(obj,context)
        httpMethod=action['method']
        uri=action['uri']
        api._attach_method(defn,pp.for_action(uri),httpMethod,execute_action)
        
    @classmethod
    def _instrument_read_method(api,defn,pp):
        # respond to specific object GETs
        def read_object(context):
            obj=hyperdns_context.get('context')
            if obj==None:
                return result.not_found()
            return HAL.hal_object(
                data=obj,
                scalars=defn.scalars,
                collections=defn.collections,
                properties=defn.properties
                )
        api._attach_method(defn,pp,'GET',read_object)

    @classmethod
    def _instrument_write_or_create_method(api,defn,pp):
        # respond to POSTs to create objects
        def update_or_create_object(context):
            obj=hyperdns_context.get('context')
            if obj==None:
                if defn.create==None:
                    return result.forbidden()
                obj=defn.create(context)
            else:
                if defn.update==None:
                    return result.forbidden()
                obj=defn.update(obj,context)
            if isinstance(obj,Response):
                return obj
            return HAL.hal_object(
                data=obj,
                scalars=defn.scalars,
                collections=defn.collections,
                properties=defn.properties
                )
        api._attach_method(defn,pp,'POST',update_or_create_object)

    @classmethod
    def _instrument_delete_method(api,defn,pp):
        # respond to specific object GETs
        def delete_object(context):
            obj=hyperdns_context.get('context')
            if obj==None:
                return result.not_found()
            defn.delete(obj,context)
            return result.ok()
        api._attach_method(defn,pp,'DELETE',delete_object)
        
    @classmethod
    def _instrument_from_defn_with_pp(api,defn,pp):
        """Instrument a definition, this is NOT called recursviely.
        """
        # support 'GET'
        api._instrument_read_method(defn,pp)
        # support 'POST'
        if defn.create!=None or defn.update!=None:
            api._instrument_write_or_create_method(defn,pp)
        # support 'DELETE'
        if defn.delete!=None:
            api._instrument_delete_method(defn,pp)
            
        # implement actions
        for action in defn.actions:
            api._instrument_action(defn,pp,action)
            
            
    @classmethod
    def _attach_method(api,defn,pp,httpMethod,func,returns='hal'):

        urlparts = urlparse(pp.uri)
        urlparts_list = list(urlparts)
        uri=urlparts_list[2]
        
        # find the returns
        if returns=='hal':
            function=result.returns_hal(func)
        elif returns=='json':
            function=result.returns_json(func)
        else:
            raise Exception('did not recognize return type')

        # we must mangle the names to avoid:
        #   "View function mapping is overwriting an existing endpoint function: core_action"
        mname=uri.replace(
            "<","_").replace(
            ">","_").replace(
            ":","_").replace(
            "/","_")+httpMethod

                            
        # this is what binds the existing environment onto the context.
        # this is where the dispatch thread enters the picture
        # this gets assigned to the class as an instance method, and it
        # is called on the instance of context or last_context
        def process_object(*args,**kwargs):
            return function(pp.setup_context(kwargs,request))

        process_object.__name__=mname

        # this calls process_object
        api_func=hyperdns.flask._app.route(uri,methods=[httpMethod])(
            api.catch_errors(
                hyperdns.flask.crossdomain(origin='*',
                    headers=['Authorization','X-Requested-With', 'Content-Type', 'Origin'])(
                    hyperdns.flask.auth(mustHaveRole=None)(
                        hyperdns.flask.db_commit()(
                            process_object
                )))))
        setattr(defn.class_,mname,api_func)   
   
    @staticmethod
    def catch_errors(f):
        @wraps(f)
        def _catch_errors(*args,**kwargs):
            try:
                r = f(*args, **kwargs)
            except Exception as E:
                traceback.print_exc()
                r=result.fail(E)
            return r
        return _catch_errors
        
   
