"""This includes CORS

"""
import json
import datetime
from functools import update_wrapper
from flask import request,make_response,_request_ctx_stack,Response
import jwt
import hyperdns.flask

#
#
def crossdomain(origin=None, methods=None, headers=None,max_age=21600, attach_to_all=True,automatic_options=True):
    """ Wrapper functions for CORS """
    if methods is not None:
        methods = ', '.join(sorted(x.upper() for x in methods))
    if headers is not None and not isinstance(headers, str):
        headers = ', '.join(x.upper() for x in headers)
    if not isinstance(origin, str):
        origin = ', '.join(origin)
    if isinstance(max_age, datetime.timedelta):
        max_age = max_age.total_seconds()
    
    def get_methods():
        if methods is not None:
            return methods
        
        options_resp = hyperdns.flask._app.make_default_options_response()
        return options_resp.headers['allow']
    
    def decorator(f):
        def _crossdomain(*args, **kwargs):
            if automatic_options and request.method == 'OPTIONS':
                resp = _app.make_default_options_response()
            else:
                resp = make_response(f(*args, **kwargs))
            if not attach_to_all and request.method != 'OPTIONS':
                return resp
            h = resp.headers
            h['Access-Control-Allow-Origin'] = origin
            h['Access-Control-Allow-Methods'] = get_methods()
            h['Access-Control-Max-Age'] = str(max_age)
            if headers is not None:
                h['Access-Control-Allow-Headers'] = headers
            return resp
        
        f.provide_automatic_options = False
        f.required_methods = ['OPTIONS']
        return update_wrapper(_crossdomain, f)
    return decorator



# primary authentication object
class auth(object):
    """
    from:
        http://www.vinaysahni.com/best-practices-for-a-pragmatic-restful-api#json-requests
        http://broadcast.oreilly.com/2011/06/the-good-the-bad-the-ugly-of-rest-apis.html
        http://blog.auth0.com/2014/01/07/angularjs-authentication-with-cookies-vs-token/
    
    HTTPS/SSL
    
    SSL is required at all times.  Files served to the client are public, and can be delivered
    via HTTP from public sources like CDNs or S3.  Any data that is not in the public domain
    must be protected, minimally, by SSL.
    
    
    Authentication
    
    A RESTful API should be stateless. This means that request authentication should not
    depend on cookies or sessions. Instead, each request should come with some sort
    authentication credentials.  Every request should be authenticated and authorized.
    Revocation must be available in real-time via a blacklist.
    
    REST api credentials should not directly involve the principal credentials.  Instead,
    an API consumer should request an access token with a fixed viability duration.  Periodically
    this token must be refreshed using the principal tokens.
    
    Authentication data should arrive via headers and should itself be considered public.
    HTTPS w/ Basic authentication has the problem of passing principal credentials on every
    request and requiring direct validation against the primary principal verification database.
    Use of an ancillary token minimizes the exposure of principal credentials and decouples
    the primary verification mechanism from each request while retaining the REST requirement
    of per-request authentication and authorization.
    
    """
    DISABLE_AUTH=False
    CLIENT_BASE=""
    CLIENT_ID = ""
    CLIENT_SECRET = ""
    INTERNAL_SECRET = ""

    def __init__(self,mustHaveRole=None):
        self.mustHaveRole=mustHaveRole
    
    @classmethod
    def config(cls,config):
        assert config!=None
        cvars=['DISABLE_AUTH',
                'CLIENT_BASE',
                'CLIENT_ID',
                'CLIENT_SECRET',
                'INTERNAL_SECRET']
        for prop in cvars:
            setattr(cls,prop,config.get(prop,getattr(cls,prop)))
        
          
    @classmethod
    def _load_user_and_roots(cls,jwt_payload,request):
        raise Exception('Use @web.auth.user_loader annotation somewhere')
    
    def __call__(self,f):
        cls=self.__class__
        
        def _extract_payload_from_jwt():
            auth = request.headers.get('Authorization', None)
            if not auth:
              return cls.require({'code': 'authorization_header_missing', 'description': 'Authorization header is expected'})
            
            parts = auth.split()                
            if parts[0].lower() != 'bearer':
              return cls.require({'code': 'invalid_header', 'description': 'Authorization header must start with Bearer'})
            elif len(parts) == 1:
              return cls.require({'code': 'invalid_header', 'description': 'Token not found'})
            elif len(parts) > 2:
              return cls.require({'code': 'invalid_header', 'description': 'Authorization header must be Bearer + \s + token'})

            
            token = parts[1]
            try:
                if token=='my-jwt':
                    payload={
                        'aud':cls.CLIENT_ID
                    }
                else:
                    #raw=base64.b64decode(INTERNAL_SECRET.replace("_","/").replace("-","+"))
                    payload = jwt.decode(
                        token,
                        cls.INTERNAL_SECRET
                    )
            except jwt.ExpiredSignature:
                return cls.require({'code': 'token_expired', 'description': 'token is expired'})
            except jwt.DecodeError:
                return cls.require({'code': 'token_invalid_signature', 'description': 'token signature is invalid'})
            
            if payload['aud'] != cls.CLIENT_ID:
              return cls.require({'code': 'invalid_audience', 'description': 'the audience does not match. expected: ' + cls.CLIENT_ID})
            return (jwt,payload)
                 
                       
        def _manage_authentication(*args, **kwargs):

            # do not require authentication for 'OPTIONS' requests
            # can we truncate here?
            if request.method=='OPTIONS':
                return None
            else:
                if cls.DISABLE_AUTH:
                    (token,payload) = ('my-jwt','my-payload')
                    (user,roots) = cls._load_user_and_roots(None,request)
                else:
                    (token,payload)=_extract_payload_from_jwt()
                    (user,roots)=self.__class__._load_user_and_roots(payload,request) 

                if user is None:
                    raise Exception('User does not exist')
                
                # available as request.current_user
                _request_ctx_stack.top.current_user = user
                
                _request_ctx_stack.top.hyperdns = {
                    'user':user,
                    'jwt':token,
                    #'account':user.account,
                    'available_roots':roots
                }
                                
                return f(*args, **kwargs)
        d=_manage_authentication
        d.__name__=f.__name__
        return d
    
    @classmethod
    def require(cls,msg):
        """Return a 401 response and a JSON explaination of the refusal
        """
        print(request.headers)
        return Response(json.dumps(msg), mimetype='application/json',status=401)
    
    @classmethod
    def user_loader(cls,callback):
        """Decorator to tag a callback as the function returning the user and
        the roots
        """
        cls._load_user_and_roots=staticmethod(callback)
        return callback
        
    
    @classmethod
    def generate_internal_key(cls,user):
        """Redirect back to the main app"""  
        import hyperdns.flask.result as result  
        token=jwt.encode({
           "user_id": "%s" % user.id,
           "aud":cls.CLIENT_ID,
           }, cls.INTERNAL_SECRET)
        return token
            
#
