#
"""Utilities for returning results
"""
import io
import json
import gzip
import datetime
from flask import Response,make_response,request
from functools import wraps
import hyperdns.hal.render as HAL
from hyperdns.flask import hyperdns_context


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            ARGS = ('year', 'month', 'day', 'hour', 'minute',
                     'second', 'microsecond')
            return {'__type__': 'datetime.datetime',
                    'args': [getattr(obj, a) for a in ARGS]}
        elif isinstance(obj, datetime.date):
            ARGS = ('year', 'month', 'day')
            return {'__type__': 'datetime.date',
                    'args': [getattr(obj, a) for a in ARGS]}
        elif isinstance(obj, datetime.time):
            ARGS = ('hour', 'minute', 'second', 'microsecond')
            return {'__type__': 'datetime.time',
                    'args': [getattr(obj, a) for a in ARGS]}
        elif isinstance(obj, datetime.timedelta):
            ARGS = ('days', 'seconds', 'microseconds')
            return {'__type__': 'datetime.timedelta',
                    'args': [getattr(obj, a) for a in ARGS]}
        else:
            return super().default(obj)


class EnhancedJSONDecoder(json.JSONDecoder):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, object_hook=self.object_hook,
                         **kwargs)

    def object_hook(self, d): 
        if '__type__' not in d:
            return d
        o = sys.modules[__name__]
        for e in d['__type__'].split('.'):
            o = getattr(o, e)
        args, kwargs = d.get('args', ()), d.get('kwargs', {})
        return o(*args, **kwargs)


class result(object):
    """
    HTTP status codes
    
    HTTP defines a bunch of meaningful status codes that can be returned from your API. These can be leveraged to help the API consumers route their responses accordingly. I've curated a short list of the ones that you definitely should be using:
        
        200 OK - Response to a successful GET, PUT, PATCH or DELETE. Can also be used for a POST that doesn't result in a creation.
        201 Created - Response to a POST that results in a creation. Should be combined with a Location header pointing to the location of the new resource
        204 No Content - Response to a successful request that won't be returning a body (like a DELETE request)
        304 Not Modified - Used when HTTP caching headers are in play
        400 Bad Request - The request is malformed, such as if the body does not parse
        401 Unauthorized - When no or invalid authentication details are provided. Also useful to trigger an auth popup if the API is used from a browser
        403 Forbidden - When authentication succeeded but authenticated user doesn't have access to the resource
        404 Not Found - When a non-existent resource is requested
        405 Method Not Allowed - When an HTTP method is being requested that isn't allowed for the authenticated user
        410 Gone - Indicates that the resource at this end point is no longer available. Useful as a blanket response for old API versions
        415 Unsupported Media Type - If incorrect content type was provided as part of the request
        422 Unprocessable Entity - Used for validation errors
        429 Too Many Requests - When a request is rejected due to rate limiting
    """
    @classmethod
    def ok(cls):
        return make_response('{"status":"ok"}',200)
    @classmethod
    def fail(cls,e=''):
        return make_response('{"status":"error","message":"%s"}' % e,500)
    @classmethod
    def plaintext(cls,txt):
        return Response(txt, mimetype='text/plain')
    @classmethod
    def json(cls,obj):
        return HAL.hal_json(data=obj)
        
    @classmethod
    def created(cls,obj):
        """
        201 Created - Response to a POST that results in a creation. Should be combined with a Location header pointing to the location of the
        """
        return make_response('',201)
    
    @classmethod
    def noContent(cls):
        """
        204 No Content - Response to a successful request that won't be returning a body (like a DELETE request)
        """
        return make_response('',204)
    
    @classmethod
    def forbidden(cls):
        """
        403 Forbidden - When authentication succeeded but authenticated user doesn't have access to the resource
        """
        return make_response('',403)
    
    @classmethod
    def bad_request(cls):
        return make_response('',400)
    
    @classmethod
    def not_found(cls):
        return make_response('',404)
    
    @classmethod
    def gone(cls,obj):
        """
        410 Gone - Indicates that the resource at this end point is no longer available. Useful as a blanket response for old API versions
        """
        return make_response('',410)
    
    @classmethod
    def unsupportedMediaType(cls):
        """
        415 Unsupported Media Type - If incorrect content type was provided as part of the request
        """
        return make_response('',415)
    
    @classmethod
    def tooManyRequests(cls,obj):
        """
        429 Too Many Requests - When a request is rejected due to rate limiting
        """
        return make_response('',429)

    @staticmethod
    def returns_json(f):
        """Indicates to flask that we're returning HAL.
    
        We're using the HAL specification by Mike Kelley
            http://stateless.co/hal_specification.html
    
        Design Decisions:
            Pretty Print By Default
                We choose to pretty print by default and apply compression to the response.
                We believe that this is an optimal balance between the ability for humans
                to rapidly and visually inspect response JSON against any increase in size
                or computing complexity.  Responses should be sent compressed via the
                standard accept-encoding mechanism.
        
            Compression
                We compress the resulting json if the client accepts gzip encoding.  The
                compression logic is only called when there is a data payload to return and
                when the data length is greater than 256 bytes.
            
                see: http://ramisayar.com/howto-automatically-enable-gzip-compression-on-aws-elastic-beanstalk/
    
        Example
        @web.returns_json
        def myMethod(...):
            return {}
        """
        @wraps(f)
        def _returns_json(*args, **kwargs):
            r = f(*args, **kwargs)
            # pass through any responses that are already formed
            if isinstance(r,Response):
                return r
            else:
                # pretty print the json and return a json object
                jsonResult=json.dumps(r,sort_keys=True, indent=4,
                    separators=(',', ': '))
                return Response(jsonResult, mimetype='application/json')
    
        return compress(_returns_json)

    
    @staticmethod
    def returns_hal(f):
        """Indicates to flask that we're returning HAL.
    
        We're using the HAL specification by Mike Kelley
            http://stateless.co/hal_specification.html
    
        Design Decisions:
            Pretty Print By Default
                We choose to pretty print by default and apply compression to the response.
                We believe that this is an optimal balance between the ability for humans
                to rapidly and visually inspect response JSON against any increase in size
                or computing complexity.  Responses should be sent compressed via the
                standard accept-encoding mechanism.
        
            Compression
                We compress the resulting json if the client accepts gzip encoding.  The
                compression logic is only called when there is a data payload to return and
                when the data length is greater than 256 bytes.
            
                see: http://ramisayar.com/howto-automatically-enable-gzip-compression-on-aws-elastic-beanstalk/
    
        Example
        @web.returns_json
        def myMethod(...):
            return {}
        """
        @wraps(f)
        def _returns_hal(*args, **kwargs):
            r = f(*args, **kwargs)
            # pass through any responses that are already formed
            if isinstance(r,Response):
                return r
            else:
                if isinstance(r,list):
                    r=HAL.hal_array(data=r,name="data")
                hal=r._hal(request.url)
                jsonResult=json.dumps(hal,sort_keys=True,
                    cls=EnhancedJSONEncoder,
                     indent=4,separators=(',', ': '))
                return Response(jsonResult, mimetype='application/hal+json;version=0')
    
        return compress(_returns_hal)
        
        
# function annotation - example: @compress, @returns_json, @returns_hal
def compress(f):
    """Indicates to flask that we're returning HAL.
    
    We're using the HAL specification by Mike Kelley
        http://stateless.co/hal_specification.html
    
    Design Decisions:
        Pretty Print By Default
            We choose to pretty print by default and apply compression to the response.
            We believe that this is an optimal balance between the ability for humans
            to rapidly and visually inspect response JSON against any increase in size
            or computing complexity.  Responses should be sent compressed via the
            standard accept-encoding mechanism.
        
        Compression
            We compress the resulting json if the client accepts gzip encoding.  The
            compression logic is only called when there is a data payload to return and
            when the data length is greater than 256 bytes.
            
            see: http://ramisayar.com/howto-automatically-enable-gzip-compression-on-aws-elastic-beanstalk/
    
    Example
    @web.returns_json
    def myMethod(...):
        return {}
    """
    @wraps(f)
    def _compress(*args, **kwargs):
        response = f(*args, **kwargs)
        # pass through any responses that are already formed
        assert isinstance(response,Response)
        
        if len(response.data)>256:
            # check to see if we need to compress
            accept_encoding = request.headers.get('Accept-Encoding', '')
            
            if 'gzip' in accept_encoding.lower():
                response.direct_passthrough = False
                
                gzip_buffer = io.BytesIO()
                gzip_file = gzip.GzipFile(mode='wb',
                                          fileobj=gzip_buffer)
                gzip_file.write(response.data)
                gzip_file.close()
                
                response.data = gzip_buffer.getvalue()
                response.headers['Content-Encoding'] = 'gzip'
                response.headers['Vary'] = 'Accept-Encoding'
                response.headers['Content-Length'] = len(response.data)
            
        return response
    
    return _compress




        

